from abc import abstractmethod
from typing import List, Dict, Union

import torch
import logging

from collections import Counter
from collections import defaultdict

from segtok.segmenter import split_single
from segtok.tokenizer import split_contractions
from segtok.tokenizer import word_tokenizer


log = logging.getLogger('flair')


class Dictionary:
    """
    This class holds a dictionary that maps strings to IDs, used to generate one-hot encodings of strings.
    """

    def __init__(self, add_unk=True):
        # init dictionaries
        self.item2idx: Dict[str, int] = {}
        self.idx2item: List[str] = []

        # in order to deal with unknown tokens, add <unk>
        if add_unk:
            self.add_item('<unk>')

    def add_item(self, item: str) -> int:
        """
        add string - if already in dictionary returns its ID. if not in dictionary, it will get a new ID.
        :param item: a string for which to assign an id
        :return: ID of string
        """
        item = item.encode('utf-8')
        if item not in self.item2idx:
            self.idx2item.append(item)
            self.item2idx[item] = len(self.idx2item) - 1
        return self.item2idx[item]

    def get_idx_for_item(self, item: str) -> int:
        """
        returns the ID of the string, otherwise 0
        :param item: string for which ID is requested
        :return: ID of string, otherwise 0
        """
        item = item.encode('utf-8')
        if item in self.item2idx.keys():
            return self.item2idx[item]
        else:
            return 0

    def get_items(self) -> List[str]:
        items = []
        for item in self.idx2item:
            items.append(item.decode('UTF-8'))
        return items

    def __len__(self) -> int:
        return len(self.idx2item)

    def get_item_for_index(self, idx):
        return self.idx2item[idx].decode('UTF-8')

    def save(self, savefile):
        import pickle
        with open(savefile, 'wb') as f:
            mappings = {
                'idx2item': self.idx2item,
                'item2idx': self.item2idx
            }
            pickle.dump(mappings, f)

    @classmethod
    def load_from_file(cls, filename: str):
        import pickle
        dictionary: Dictionary = Dictionary()
        with open(filename, 'rb') as f:
            mappings = pickle.load(f, encoding='latin1')
            idx2item = mappings['idx2item']
            item2idx = mappings['item2idx']
            dictionary.item2idx = item2idx
            dictionary.idx2item = idx2item
        return dictionary

    @classmethod
    def load(cls, name: str):
        from flair.file_utils import cached_path
        if name == 'chars' or name == 'common-chars':
            base_path = 'https://s3.eu-central-1.amazonaws.com/alan-nlp/resources/models/common_characters'
            char_dict = cached_path(base_path, cache_dir='datasets')
            return Dictionary.load_from_file(char_dict)

        return Dictionary.load_from_file(name)


class Label:
    """
    This class represents a label of a sentence. Each label has a value and optionally a confidence score. The
    score needs to be between 0.0 and 1.0. Default value for the score is 1.0.
    """

    def __init__(self, value: str, score: float = 1.0):
        self.value = value
        self.score = score
        super().__init__()

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if not value and value != '':
            raise ValueError('Incorrect label value provided. Label value needs to be set.')
        else:
            self._value = value

    @property
    def score(self):
        return self._score

    @score.setter
    def score(self, score):
        if 0.0 <= score <= 1.0:
            self._score = score
        else:
            self._score = 1.0

    def to_dict(self):
        return {
            'value': self.value,
            'confidence': self.score
        }

    def __str__(self):
        return "{} ({})".format(self._value, self._score)

    def __repr__(self):
        return "{} ({})".format(self._value, self._score)


class Token:
    """
    This class represents one word in a tokenized sentence. Each token may have any number of tags. It may also point
    to its head in a dependency tree.
    """

    def __init__(self,
                 text: str,
                 idx: int = None,
                 head_id: int = None,
                 whitespace_after: bool = True,
                 start_position: int = None
                 ):
        self.text: str = text
        self.idx: int = idx
        self.head_id: int = head_id
        self.whitespace_after: bool = whitespace_after

        self.start_pos = start_position
        self.end_pos = start_position + len(text) if start_position is not None else None

        self.sentence: Sentence = None
        self._embeddings: Dict = {}
        self.tags: Dict[str, Label] = {}

    def add_tag_label(self, tag_type: str, tag: Label):
        self.tags[tag_type] = tag

    def add_tag(self, tag_type: str, tag_value: str, confidence=1.0):
        tag = Label(tag_value, confidence)
        self.tags[tag_type] = tag

    def get_tag(self, tag_type: str) -> Label:
        if tag_type in self.tags: return self.tags[tag_type]
        return Label('')

    def get_head(self):
        return self.sentence.get_token(self.head_id)

    def set_embedding(self, name: str, vector: torch.autograd.Variable):
        self._embeddings[name] = vector.cpu()

    def clear_embeddings(self):
        self._embeddings: Dict = {}

    def get_embedding(self) -> torch.FloatTensor:

        embeddings = [self._embeddings[embed] for embed in sorted(self._embeddings.keys())]

        if embeddings:
            return torch.cat(embeddings, dim=0)

        return torch.FloatTensor()

    @property
    def start_position(self) -> int:
        return self.start_pos

    @property
    def end_position(self) -> int:
        return self.end_pos

    @property
    def embedding(self):
        return self.get_embedding()

    def __str__(self) -> str:
        return 'Token: {} {}'.format(self.idx, self.text) if self.idx is not None else 'Token: {}'.format(self.text)

    def __repr__(self) -> str:
        return 'Token: {} {}'.format(self.idx, self.text) if self.idx is not None else 'Token: {}'.format(self.text)


class Span:
    """
    This class represents one textual span consisting of Tokens. A span may have a tag.
    """

    def __init__(self, tokens: List[Token], tag: str = None, score=1.):
        self.tokens = tokens
        self.tag = tag
        self.score = score
        self.start_pos = None
        self.end_pos = None

        if tokens:
            self.start_pos = tokens[0].start_position
            self.end_pos = tokens[len(tokens) - 1].end_position

    @property
    def text(self) -> str:
        return ' '.join([t.text for t in self.tokens])

    def to_original_text(self) -> str:
        str = ''
        pos = self.tokens[0].start_pos
        for t in self.tokens:
            while t.start_pos != pos:
                str += ' '
                pos += 1

            str += t.text
            pos += len(t.text)

        return str

    def to_dict(self):
        return {
            'text': self.to_original_text(),
            'start_pos': self.start_pos,
            'end_pos': self.end_pos,
            'type': self.tag,
            'confidence': self.score
        }

    def __str__(self) -> str:
        ids = ','.join([str(t.idx) for t in self.tokens])
        return '{}-span [{}]: "{}"'.format(self.tag, ids, self.text) \
            if self.tag is not None else 'span [{}]: "{}"'.format(ids, self.text)

    def __repr__(self) -> str:
        ids = ','.join([str(t.idx) for t in self.tokens])
        return '<{}-span ({}): "{}">'.format(self.tag, ids, self.text) \
            if self.tag is not None else '<span ({}): "{}">'.format(ids, self.text)


class Sentence:
    """
    A Sentence is a list of Tokens and is used to represent a sentence or text fragment.
    """

    def __init__(self, text: str = None, use_tokenizer: bool = False, labels: Union[List[Label], List[str]] = None):

        super(Sentence, self).__init__()

        self.tokens: List[Token] = []

        self.labels: List[Label] = []
        if labels is not None: self.add_labels(labels)

        self._embeddings: Dict = {}

        # if text is passed, instantiate sentence with tokens (words)
        if text is not None:

            # tokenize the text first if option selected
            if use_tokenizer:

                # use segtok for tokenization
                tokens = []
                sentences = split_single(text)
                for sentence in sentences:
                    contractions = split_contractions(word_tokenizer(sentence))
                    tokens.extend(contractions)

                # determine offsets for whitespace_after field
                index = text.index
                running_offset = 0
                last_word_offset = -1
                last_token = None
                for word in tokens:
                    try:
                        word_offset = index(word, running_offset)
                        start_position = word_offset
                    except:
                        word_offset = last_word_offset + 1
                        start_position = running_offset + 1 if running_offset > 0 else running_offset

                    token = Token(word, start_position=start_position)
                    self.add_token(token)

                    if word_offset - 1 == last_word_offset and last_token is not None:
                        last_token.whitespace_after = False

                    word_len = len(word)
                    running_offset = word_offset + word_len
                    last_word_offset = running_offset - 1
                    last_token = token

            # otherwise assumes whitespace tokenized text
            else:
                # add each word in tokenized string as Token object to Sentence
                word = ''
                for index, char in enumerate(text):
                    if char == ' ':
                        if len(word) > 0:
                            token = Token(word, start_position=index-len(word))
                            self.add_token(token)

                        word = ''
                    else:
                        word += char
                # increment for last token in sentence if not followed by whtespace
                index += 1
                if len(word) > 0:
                    token = Token(word, start_position=index-len(word))
                    self.add_token(token)

    def get_token(self, token_id: int) -> Token:
        for token in self.tokens:
            if token.idx == token_id:
                return token

    def add_token(self, token: Token):
        self.tokens.append(token)

        # set token idx if not set
        token.sentence = self
        if token.idx is None:
            token.idx = len(self.tokens)

    def get_spans(self, tag_type: str, min_score=-1) -> List[Span]:

        spans: List[Span] = []

        current_span = []

        tags = defaultdict(lambda: 0.0)

        previous_tag_value: str = 'O'
        for token in self:

            tag: Label = token.get_tag(tag_type)
            tag_value = tag.value

            # non-set tags are OUT tags
            if tag_value == '' or tag_value == 'O':
                tag_value = 'O-'

            # anything that is not a BIOES tag is a SINGLE tag
            if tag_value[0:2] not in ['B-', 'I-', 'O-', 'E-', 'S-']:
                tag_value = 'S-' + tag_value

            # anything that is not OUT is IN
            in_span = False
            if tag_value[0:2] not in ['O-']:
                in_span = True

            # single and begin tags start a new span
            starts_new_span = False
            if tag_value[0:2] in ['B-', 'S-']:
                starts_new_span = True

            if previous_tag_value[0:2] in ['S-'] and previous_tag_value[2:] != tag_value[2:] and in_span:
                starts_new_span = True

            if (starts_new_span or not in_span) and len(current_span) > 0:
                scores = [t.get_tag(tag_type).score for t in current_span]
                span_score = sum(scores) / len(scores)
                if span_score > min_score:
                    spans.append(Span(
                        current_span,
                        tag=sorted(tags.items(), key=lambda k_v: k_v[1], reverse=True)[0][0],
                        score=span_score)
                    )
                current_span = []
                tags = defaultdict(lambda: 0.0)

            if in_span:
                current_span.append(token)
                weight = 1.1 if starts_new_span else 1.0
                tags[tag_value[2:]] += weight

            # remember previous tag
            previous_tag_value = tag_value

        if len(current_span) > 0:
            scores = [t.get_tag(tag_type).score for t in current_span]
            span_score = sum(scores) / len(scores)
            if span_score > min_score:
                spans.append(Span(
                    current_span,
                    tag=sorted(tags.items(), key=lambda k_v: k_v[1], reverse=True)[0][0],
                    score=span_score)
                )

        return spans

    def add_label(self, label: Union[Label, str]):
        if type(label) is Label:
            self.labels.append(label)

        elif type(label) is str:
            self.labels.append(Label(label))

    def add_labels(self, labels: Union[List[Label], List[str]]):
        for label in labels:
            self.add_label(label)

    def get_label_names(self) -> List[str]:
        return [label.value for label in self.labels]

    @property
    def embedding(self):
        return self.get_embedding()

    def set_embedding(self, name: str, vector):
        self._embeddings[name] = vector.cpu()

    def get_embedding(self) -> torch.autograd.Variable:
        embeddings = []
        for embed in sorted(self._embeddings.keys()):
            embedding = self._embeddings[embed]
            embeddings.append(embedding)

        if embeddings:
            return torch.cat(embeddings, dim=0)

        return torch.FloatTensor()

    def clear_embeddings(self, also_clear_word_embeddings: bool = True):
        self._embeddings: Dict = {}

        if also_clear_word_embeddings:
            for token in self:
                token.clear_embeddings()

    def cpu_embeddings(self):
        for name, vector in self._embeddings.items():
            self._embeddings[name] = vector.cpu()

    def to_tagged_string(self, main_tag=None) -> str:
        list = []
        for token in self.tokens:
            list.append(token.text)

            tags: List[str] = []
            for tag_type in token.tags.keys():

                if main_tag is not None and main_tag != tag_type: continue

                if token.get_tag(tag_type).value == '' or token.get_tag(tag_type).value == 'O': continue
                tags.append(token.get_tag(tag_type).value)
            all_tags = '<' + '/'.join(tags) + '>'
            if all_tags != '<>':
                list.append(all_tags)
        return ' '.join(list)

    def to_tokenized_string(self) -> str:
        return ' '.join([t.text for t in self.tokens])

    def to_plain_string(self):
        plain = ''
        for token in self.tokens:
            plain += token.text
            if token.whitespace_after: plain += ' '
        return plain.rstrip()

    def convert_tag_scheme(self, tag_type: str = 'ner', target_scheme: str = 'iob'):

        tags: List[Label] = []
        for token in self.tokens:
            token: Token = token
            tags.append(token.get_tag(tag_type))

        if target_scheme == 'iob':
            iob2(tags)

        if target_scheme == 'iobes':
            iob2(tags)
            tags = iob_iobes(tags)

        for index, tag in enumerate(tags):
            self.tokens[index].add_tag(tag_type, tag)

    def infer_space_after(self):
        """
        Heuristics in case you wish to infer whitespace_after values for tokenized text. This is useful for some old NLP
        tasks (such as CoNLL-03 and CoNLL-2000) that provide only tokenized data with no info of original whitespacing.
        :return:
        """
        last_token = None
        quote_count: int = 0
        # infer whitespace after field

        for token in self.tokens:
            if token.text == '"':
                quote_count += 1
                if quote_count % 2 != 0:
                    token.whitespace_after = False
                elif last_token is not None:
                    last_token.whitespace_after = False

            if last_token is not None:

                if token.text in ['.', ':', ',', ';', ')', 'n\'t', '!', '?']:
                    last_token.whitespace_after = False

                if token.text.startswith('\''):
                    last_token.whitespace_after = False

            if token.text in ['(']:
                token.whitespace_after = False

            last_token = token
        return self

    def to_original_text(self) -> str:
        str = ''
        pos = 0
        for t in self.tokens:
            while t.start_pos != pos:
                str += ' '
                pos += 1

            str += t.text
            pos += len(t.text)

        return str

    def to_dict(self, tag_type: str = None):
        labels = []
        entities = []

        if tag_type:
            entities = [span.to_dict() for span in self.get_spans(tag_type)]
        if self.labels:
            labels = [l.to_dict() for l in self.labels]

        return {
            'text': self.to_original_text(),
            'labels': labels,
            'entities': entities
        }

    def __getitem__(self, idx: int) -> Token:
        return self.tokens[idx]

    def __iter__(self):
        return iter(self.tokens)

    def __repr__(self):
        return 'Sentence: "{}" - {} Tokens'.format(' '.join([t.text for t in self.tokens]), len(self))

    def __copy__(self):
        s = Sentence()
        for token in self.tokens:
            nt = Token(token.text)
            for tag_type in token.tags:
                nt.add_tag(tag_type, token.get_tag(tag_type).value, token.get_tag(tag_type).score)

            s.add_token(nt)
        return s

    def __str__(self) -> str:
        return 'Sentence: "{}" - {} Tokens'.format(' '.join([t.text for t in self.tokens]), len(self))

    def __len__(self) -> int:
        return len(self.tokens)


class Corpus:

    @property
    @abstractmethod
    def train(self) -> List[Sentence]:
        pass

    @property
    @abstractmethod
    def dev(self) -> List[Sentence]:
        pass

    @property
    @abstractmethod
    def test(self) -> List[Sentence]:
        pass

    @abstractmethod
    def downsample(self, percentage: float = 0.1, only_downsample_train=False):
        """Downsamples this corpus to a percentage of the sentences."""
        pass

    @abstractmethod
    def get_all_sentences(self) -> List[Sentence]:
        """Gets all sentences in the corpus (train, dev and test splits together)."""
        pass

    @abstractmethod
    def make_tag_dictionary(self, tag_type: str) -> Dictionary:
        """Produces a dictionary of token tags of tag_type."""
        pass

    @abstractmethod
    def make_label_dictionary(self) -> Dictionary:
        """
        Creates a dictionary of all labels assigned to the sentences in the corpus.
        :return: dictionary of labels
        """
        pass


class TaggedCorpus(Corpus):
    def __init__(self, train: List[Sentence], dev: List[Sentence], test: List[Sentence], name: str = 'corpus'):
        self._train: List[Sentence] = train
        self._dev: List[Sentence] = dev
        self._test: List[Sentence] = test
        self.name: str = name

    @property
    def train(self) -> List[Sentence]:
        return self._train

    @property
    def dev(self) -> List[Sentence]:
        return self._dev

    @property
    def test(self) -> List[Sentence]:
        return self._test

    def downsample(self, percentage: float = 0.1, only_downsample_train=False):

        self._train = self._downsample_to_proportion(self.train, percentage)
        if not only_downsample_train:
            self._dev = self._downsample_to_proportion(self.dev, percentage)
            self._test = self._downsample_to_proportion(self.test, percentage)

        return self

    def get_all_sentences(self) -> List[Sentence]:
        all_sentences: List[Sentence] = []
        all_sentences.extend(self.train)
        all_sentences.extend(self.dev)
        all_sentences.extend(self.test)
        return all_sentences

    def make_tag_dictionary(self, tag_type: str) -> Dictionary:

        # Make the tag dictionary
        tag_dictionary: Dictionary = Dictionary()
        tag_dictionary.add_item('O')
        for sentence in self.get_all_sentences():
            for token in sentence.tokens:
                token: Token = token
                tag_dictionary.add_item(token.get_tag(tag_type).value)
        tag_dictionary.add_item('<START>')
        tag_dictionary.add_item('<STOP>')
        return tag_dictionary

    def make_label_dictionary(self) -> Dictionary:
        """
        Creates a dictionary of all labels assigned to the sentences in the corpus.
        :return: dictionary of labels
        """

        labels = set(self._get_all_label_names())

        label_dictionary: Dictionary = Dictionary(add_unk=False)
        for label in labels:
            label_dictionary.add_item(label)

        return label_dictionary

    def make_vocab_dictionary(self, max_tokens=-1, min_freq=1) -> Dictionary:
        """
        Creates a dictionary of all tokens contained in the corpus.
        By defining `max_tokens` you can set the maximum number of tokens that should be contained in the dictionary.
        If there are more than `max_tokens` tokens in the corpus, the most frequent tokens are added first.
        If `min_freq` is set the a value greater than 1 only tokens occurring more than `min_freq` times are considered
        to be added to the dictionary.
        :param max_tokens: the maximum number of tokens that should be added to the dictionary (-1 = take all tokens)
        :param min_freq: a token needs to occur at least `min_freq` times to be added to the dictionary (-1 = there is no limitation)
        :return: dictionary of tokens
        """
        tokens = self._get_most_common_tokens(max_tokens, min_freq)

        vocab_dictionary: Dictionary = Dictionary()
        for token in tokens:
            vocab_dictionary.add_item(token)

        return vocab_dictionary

    def _get_most_common_tokens(self, max_tokens, min_freq) -> List[str]:
        tokens_and_frequencies = Counter(self._get_all_tokens())
        tokens_and_frequencies = tokens_and_frequencies.most_common()

        tokens = []
        for token, freq in tokens_and_frequencies:
            if (min_freq != -1 and freq < min_freq) or (max_tokens != -1 and len(tokens) == max_tokens):
                break
            tokens.append(token)
        return tokens

    def _get_all_label_names(self) -> List[str]:
        return [label.value for sent in self.train for label in sent.labels]

    def _get_all_tokens(self) -> List[str]:
        tokens = list(map((lambda s: s.tokens), self.train))
        tokens = [token for sublist in tokens for token in sublist]
        return list(map((lambda t: t.text), tokens))

    def _downsample_to_proportion(self, list: List, proportion: float):

        counter = 0.0
        last_counter = None
        downsampled: List = []

        for item in list:
            counter += proportion
            if int(counter) != last_counter:
                downsampled.append(item)
                last_counter = int(counter)
        return downsampled

    def obtain_statistics(self, tag_type: str = None) -> dict:
        """
        Print statistics about the class distribution (only labels of sentences are taken into account) and sentence
        sizes.
        """
        return {
            "TRAIN": self._obtain_statistics_for(self.train, "TRAIN", tag_type),
            "TEST": self._obtain_statistics_for(self.test, "TEST", tag_type),
            "DEV": self._obtain_statistics_for(self.dev, "DEV", tag_type),
        }

    @staticmethod
    def _obtain_statistics_for(sentences, name, tag_type) -> dict:
        if len(sentences) == 0:
            return {}

        classes_to_count = TaggedCorpus._get_class_to_count(sentences)
        tags_to_count = TaggedCorpus._get_tag_to_count(sentences, tag_type)
        tokens_per_sentence = TaggedCorpus._get_tokens_per_sentence(sentences)

        label_size_dict = {}
        for l, c in classes_to_count.items():
            label_size_dict[l] = c

        tag_size_dict = {}
        for l, c in tags_to_count.items():
            tag_size_dict[l] = c

        return {
            'dataset': name,
            'total_number_of_documents': len(sentences),
            'number_of_documents_per_class': label_size_dict,
            'number_of_tokens_per_tag': tag_size_dict,
            'number_of_tokens': {
                'total': sum(tokens_per_sentence),
                'min': min(tokens_per_sentence),
                'max': max(tokens_per_sentence),
                'avg': sum(tokens_per_sentence) / len(sentences)
            }
        }

    @staticmethod
    def _get_tokens_per_sentence(sentences):
        return list(map(lambda x: len(x.tokens), sentences))

    @staticmethod
    def _get_class_to_count(sentences):
        class_to_count = defaultdict(lambda: 0)
        for sent in sentences:
            for label in sent.labels:
                class_to_count[label.value] += 1
        return class_to_count

    @staticmethod
    def _get_tag_to_count(sentences, tag_type):
        tag_to_count = defaultdict(lambda: 0)
        for sent in sentences:
            for word in sent.tokens:
                if tag_type in word.tags:
                    label = word.tags[tag_type]
                    tag_to_count[label.value] += 1
        return tag_to_count

    def __str__(self) -> str:
        return 'TaggedCorpus: %d train + %d dev + %d test sentences' % (len(self.train), len(self.dev), len(self.test))


def iob2(tags):
    """
    Check that tags have a valid IOB format.
    Tags in IOB1 format are converted to IOB2.
    """
    for i, tag in enumerate(tags):
        if tag.value == 'O':
            continue
        split = tag.value.split('-')
        if len(split) != 2 or split[0] not in ['I', 'B']:
            return False
        if split[0] == 'B':
            continue
        elif i == 0 or tags[i - 1].value == 'O':  # conversion IOB1 to IOB2
            tags[i].value = 'B' + tag.value[1:]
        elif tags[i - 1].value[1:] == tag.value[1:]:
            continue
        else:  # conversion IOB1 to IOB2
            tags[i].value = 'B' + tag.value[1:]
    return True


def iob_iobes(tags):
    """
    IOB -> IOBES
    """
    new_tags = []
    for i, tag in enumerate(tags):
        if tag.value == 'O':
            new_tags.append(tag.value)
        elif tag.value.split('-')[0] == 'B':
            if i + 1 != len(tags) and \
                    tags[i + 1].value.split('-')[0] == 'I':
                new_tags.append(tag.value)
            else:
                new_tags.append(tag.value.replace('B-', 'S-'))
        elif tag.value.split('-')[0] == 'I':
            if i + 1 < len(tags) and \
                    tags[i + 1].value.split('-')[0] == 'I':
                new_tags.append(tag.value)
            else:
                new_tags.append(tag.value.replace('I-', 'E-'))
        else:
            raise Exception('Invalid IOB format!')
    return new_tags


class MultiCorpus(Corpus):

    def __init__(self, corpora: List[TaggedCorpus]):
        self.corpora: List[TaggedCorpus] = corpora

    @property
    def train(self) -> List[Sentence]:
        train: List[Sentence] = []
        for corpus in self.corpora:
            train.extend(corpus.train)
        return train

    @property
    def dev(self) -> List[Sentence]:
        dev: List[Sentence] = []
        for corpus in self.corpora:
            dev.extend(corpus.dev)
        return dev

    @property
    def test(self) -> List[Sentence]:
        test: List[Sentence] = []
        for corpus in self.corpora:
            test.extend(corpus.test)
        return test

    def __str__(self):
        return '\n'.join([str(corpus) for corpus in self.corpora])

    def get_all_sentences(self) -> List[Sentence]:
        sentences = []
        for corpus in self.corpora:
            sentences.extend(corpus.get_all_sentences())
        return sentences

    def downsample(self, percentage: float = 0.1, only_downsample_train=False):

        for corpus in self.corpora:
            corpus.downsample(percentage, only_downsample_train)

        return self

    def make_tag_dictionary(self, tag_type: str) -> Dictionary:

        # Make the tag dictionary
        tag_dictionary: Dictionary = Dictionary()
        tag_dictionary.add_item('O')
        for corpus in self.corpora:
            for sentence in corpus.get_all_sentences():
                for token in sentence.tokens:
                    token: Token = token
                    tag_dictionary.add_item(token.get_tag(tag_type).value)
        tag_dictionary.add_item('<START>')
        tag_dictionary.add_item('<STOP>')
        return tag_dictionary

    def make_label_dictionary(self) -> Dictionary:

        label_dictionary: Dictionary = Dictionary(add_unk=False)
        for corpus in self.corpora:
            labels = set(corpus._get_all_label_names())

            for label in labels:
                label_dictionary.add_item(label)

        return label_dictionary



from collections import defaultdict
from pathlib import Path
from typing import Union

import numpy as np
import csv

import matplotlib
import math

matplotlib.use('Agg')

import matplotlib.pyplot as plt

# header for 'weights.txt'
WEIGHT_NAME = 1
WEIGHT_NUMBER = 2
WEIGHT_VALUE = 3


class Plotter(object):
    """
    Plots training parameters (loss, f-score, and accuracy) and training weights over time.
    Input files are the output files 'loss.tsv' and 'weights.txt' from training either a sequence tagger or text
    classification model.
    """

    @staticmethod
    def _extract_evaluation_data(file_name: Path) -> dict:
        training_curves = {
            'train': {
                'loss': [],
                'f_score': [],
                'acc': []
            },
            'test': {
                'loss': [],
                'f_score': [],
                'acc': []
            },
            'dev': {
                'loss': [],
                'f_score': [],
                'acc': []
            }
        }

        with open(file_name, 'r') as tsvin:
            tsvin = csv.reader(tsvin, delimiter='\t')

            # determine the column index of loss, f-score and accuracy for train, dev and test split
            row = next(tsvin, None)
            TRAIN_LOSS = row.index('TRAIN_LOSS')
            TRAIN_F_SCORE = row.index('TRAIN_F-SCORE')
            TRAIN_ACCURACY = row.index('TRAIN_ACCURACY')
            DEV_LOSS = row.index('DEV_LOSS')
            DEV_F_SCORE = row.index('DEV_F-SCORE')
            DEV_ACCURACY = row.index('DEV_ACCURACY')
            TEST_LOSS = row.index('TEST_LOSS')
            TEST_F_SCORE = row.index('TEST_F-SCORE')
            TEST_ACCURACY = row.index('TEST_ACCURACY')

            # then get all relevant values from the tsv
            for row in tsvin:
                if row[TRAIN_LOSS] != '_': training_curves['train']['loss'].append(float(row[TRAIN_LOSS]))
                if row[TRAIN_F_SCORE] != '_': training_curves['train']['f_score'].append(float(row[TRAIN_F_SCORE]))
                if row[TRAIN_ACCURACY] != '_': training_curves['train']['acc'].append(float(row[TRAIN_ACCURACY]))
                if row[DEV_LOSS] != '_': training_curves['dev']['loss'].append(float(row[DEV_LOSS]))
                if row[DEV_F_SCORE] != '_': training_curves['dev']['f_score'].append(float(row[DEV_F_SCORE]))
                if row[DEV_ACCURACY] != '_': training_curves['dev']['acc'].append(float(row[DEV_ACCURACY]))
                if row[TEST_LOSS] != '_': training_curves['test']['loss'].append(float(row[TEST_LOSS]))
                if row[TEST_F_SCORE] != '_': training_curves['test']['f_score'].append(float(row[TEST_F_SCORE]))
                if row[TEST_ACCURACY] != '_': training_curves['test']['acc'].append(float(row[TEST_ACCURACY]))

        return training_curves

    @staticmethod
    def _extract_weight_data(file_name: Path) -> dict:
        weights = defaultdict(lambda: defaultdict(lambda: list()))

        with open(file_name, 'r') as tsvin:
            tsvin = csv.reader(tsvin, delimiter='\t')

            for row in tsvin:
                name = row[WEIGHT_NAME]
                param = row[WEIGHT_NUMBER]
                value = float(row[WEIGHT_VALUE])

                weights[name][param].append(value)

        return weights

    @staticmethod
    def _extract_learning_rate(file_name: Path):
        lrs = []
        losses = []

        with open(file_name, 'r') as tsvin:
            tsvin = csv.reader(tsvin, delimiter='\t')
            row = next(tsvin, None)
            LEARNING_RATE = row.index('LEARNING_RATE')
            TRAIN_LOSS = row.index('TRAIN_LOSS')

            # then get all relevant values from the tsv
            for row in tsvin:
                if row[TRAIN_LOSS] != '_': losses.append(float(row[TRAIN_LOSS]))
                if row[LEARNING_RATE] != '_': lrs.append(float(row[LEARNING_RATE]))

        return lrs, losses

    def plot_weights(self, file_name: Union[str, Path]):
        if type(file_name) is str:
            file_name = Path(file_name)

        weights = self._extract_weight_data(file_name)

        total = len(weights)
        columns = 2
        rows = max(2, int(math.ceil(total / columns)))

        figsize = (5, 5)
        if rows != columns:
            figsize = (5, rows + 5)

        fig = plt.figure()
        f, axarr = plt.subplots(rows, columns, figsize=figsize)

        c = 0
        r = 0
        for name, values in weights.items():
            # plot i
            axarr[r, c].set_title(name, fontsize=6)
            for _, v in values.items():
                axarr[r, c].plot(np.arange(0, len(v)), v, linewidth=0.35)
            axarr[r, c].set_yticks([])
            axarr[r, c].set_xticks([])
            c += 1
            if c == columns:
                c = 0
                r += 1

        while r != rows and c != columns:
            axarr[r, c].set_yticks([])
            axarr[r, c].set_xticks([])
            c += 1
            if c == columns:
                c = 0
                r += 1

        # save plots
        f.subplots_adjust(hspace=0.5)
        plt.tight_layout(pad=1.0)
        path = file_name.parent / 'weights.png'
        plt.savefig(path, dpi=300)

        plt.close(fig)

    def plot_training_curves(self, file_name: Union[str, Path]):
        if type(file_name) is str:
            file_name = Path(file_name)

        fig = plt.figure(figsize=(15, 10))

        training_curves = self._extract_evaluation_data(file_name)

        # plot 1
        plt.subplot(3, 1, 1)
        if training_curves['train']['loss']:
            x = np.arange(0, len(training_curves['train']['loss']))
            plt.plot(x, training_curves['train']['loss'], label='training loss')
        if training_curves['dev']['loss']:
            x = np.arange(0, len(training_curves['dev']['loss']))
            plt.plot(x, training_curves['dev']['loss'], label='validation loss')
        if training_curves['test']['loss']:
            x = np.arange(0, len(training_curves['test']['loss']))
            plt.plot(x, training_curves['test']['loss'], label='test loss')
        plt.legend(bbox_to_anchor=(1.04, 0), loc="lower left", borderaxespad=0)
        plt.ylabel('loss')
        plt.xlabel('epochs')

        # plot 2
        plt.subplot(3, 1, 2)
        if training_curves['train']['acc']:
            x = np.arange(0, len(training_curves['train']['acc']))
            plt.plot(x, training_curves['train']['acc'], label='training accuracy')
        if training_curves['dev']['acc']:
            x = np.arange(0, len(training_curves['dev']['acc']))
            plt.plot(x, training_curves['dev']['acc'], label='validation accuracy')
        if training_curves['test']['acc']:
            x = np.arange(0, len(training_curves['test']['acc']))
            plt.plot(x, training_curves['test']['acc'], label='test accuracy')
        plt.legend(bbox_to_anchor=(1.04, 0), loc="lower left", borderaxespad=0)
        plt.ylabel('accuracy')
        plt.xlabel('epochs')

        # plot 3
        plt.subplot(3, 1, 3)
        if training_curves['train']['f_score']:
            x = np.arange(0, len(training_curves['train']['f_score']))
            plt.plot(x, training_curves['train']['f_score'], label='training f1-score')
        if training_curves['dev']['f_score']:
            x = np.arange(0, len(training_curves['dev']['f_score']))
            plt.plot(x, training_curves['dev']['f_score'], label='validation f1-score')
        if training_curves['test']['f_score']:
            x = np.arange(0, len(training_curves['test']['f_score']))
            plt.plot(x, training_curves['test']['f_score'], label='test f1-score')
        plt.legend(bbox_to_anchor=(1.04, 0), loc="lower left", borderaxespad=0)
        plt.ylabel('f1-score')
        plt.xlabel('epochs')

        # save plots
        plt.tight_layout(pad=1.0)
        path = file_name.parent / 'training.png'
        plt.savefig(path, dpi=300)

        plt.close(fig)

    def plot_learning_rate(self,
                           file_name: Union[str, Path],
                           skip_first: int = 10,
                           skip_last: int = 5):
        if type(file_name) is str:
            file_name = Path(file_name)

        lrs, losses = self._extract_learning_rate(file_name)
        lrs = lrs[skip_first:-skip_last] if skip_last > 0 else lrs[skip_first:]
        losses = losses[skip_first:-skip_last] if skip_last > 0 else losses[skip_first:]

        fig, ax = plt.subplots(1,1)
        ax.plot(lrs, losses)
        ax.set_ylabel("Loss")
        ax.set_xlabel("Learning Rate")
        ax.set_xscale('log')
        ax.xaxis.set_major_formatter(plt.FormatStrFormatter('%.0e'))

        # save plot
        plt.tight_layout(pad=1.0)
        path = file_name.parent / 'learning_rate.png'
        plt.savefig(path, dpi=300)

        plt.close(fig)



import logging
import math
from functools import partial

import torch
from torch.optim import Optimizer
from torch.optim.optimizer import required
from torch.optim.lr_scheduler import _LRScheduler, ReduceLROnPlateau


log = logging.getLogger('flair')


class SGDW(Optimizer):
    r"""Implements stochastic gradient descent (optionally with momentum) with
    weight decay from the paper `Fixing Weight Decay Regularization in Adam`_.

    Nesterov momentum is based on the formula from
    `On the importance of initialization and momentum in deep learning`__.

    Args:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float): learning rate
        momentum (float, optional): momentum factor (default: 0)
        weight_decay (float, optional): weight decay factor (default: 0)
        dampening (float, optional): dampening for momentum (default: 0)
        nesterov (bool, optional): enables Nesterov momentum (default: False)

    .. _Fixing Weight Decay Regularization in Adam:
        https://arxiv.org/abs/1711.05101

    Example:
        >>> optimizer = torch.optim.SGDW(model.parameters(), lr=0.1, momentum=0.9,
                                         weight_decay=1e-5)
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()

    __ http://www.cs.toronto.edu/%7Ehinton/absps/momentum.pdf

    .. note::
        The implementation of SGD with Momentum/Nesterov subtly differs from
        Sutskever et. al. and implementations in some other frameworks.

        Considering the specific case of Momentum, the update can be written as

        .. math::
                  v = \rho * v + g \\
                  p = p - lr * v

        where p, g, v and :math:`\rho` denote the parameters, gradient,
        velocity, and momentum respectively.

        This is in contrast to Sutskever et. al. and
        other frameworks which employ an update of the form

        .. math::
             v = \rho * v + lr * g \\
             p = p - v

        The Nesterov version is analogously modified.
    """

    def __init__(self, params, lr=required, momentum=0, dampening=0,
                 weight_decay=0, nesterov=False):
        if lr is not required and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if momentum < 0.0:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))

        defaults = dict(lr=lr, momentum=momentum, dampening=dampening,
                        weight_decay=weight_decay, nesterov=nesterov)
        if nesterov and (momentum <= 0 or dampening != 0):
            raise ValueError("Nesterov momentum requires a momentum and zero dampening")
        super(SGDW, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(SGDW, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('nesterov', False)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            weight_decay = group['weight_decay']
            momentum = group['momentum']
            dampening = group['dampening']
            nesterov = group['nesterov']

            for p in group['params']:
                if p.grad is None:
                    continue
                d_p = p.grad.data

                if momentum != 0:
                    param_state = self.state[p]
                    if 'momentum_buffer' not in param_state:
                        buf = param_state['momentum_buffer'] = torch.zeros_like(p.data)
                        buf.mul_(momentum).add_(d_p)
                    else:
                        buf = param_state['momentum_buffer']
                        buf.mul_(momentum).add_(1 - dampening, d_p)
                    if nesterov:
                        d_p = d_p.add(momentum, buf)
                    else:
                        d_p = buf

                if weight_decay != 0:
                    p.data.add_(-weight_decay, p.data)

                p.data.add_(-group['lr'], d_p)

        return loss


class AdamW(Optimizer):
    r"""Implements AdamW optimizer.

    Adam has been proposed in `Adam\: A Method for Stochastic Optimization`_.
    AdamW uses the weight decay method from the paper
    `Fixing Weight Decay Regularization in Adam`_.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay factor (default: 0)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper `On the Convergence of Adam and Beyond`_
            (default: False)

    .. _Adam\: A Method for Stochastic Optimization:
        https://arxiv.org/abs/1412.6980
    .. _Fixing Weight Decay Regularization in Adam:
        https://arxiv.org/abs/1711.05101
    .. _On the Convergence of Adam and Beyond:
        https://openreview.net/forum?id=ryQu7f-RZ
    """

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8,
                 weight_decay=0, amsgrad=False):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        defaults = dict(lr=lr, betas=betas, eps=eps,
                        weight_decay=weight_decay, amsgrad=amsgrad)
        super(AdamW, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(AdamW, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('amsgrad', False)

    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('Adam does not support sparse gradients, please consider SparseAdam instead')
                amsgrad = group['amsgrad']

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                    if amsgrad:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state['max_exp_avg_sq'] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                if amsgrad:
                    max_exp_avg_sq = state['max_exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                if amsgrad:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = max_exp_avg_sq.sqrt().add_(group['eps'])
                else:
                    denom = exp_avg_sq.sqrt().add_(group['eps'])

                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']
                step_size = group['lr'] * math.sqrt(bias_correction2) / bias_correction1

                if group['weight_decay'] != 0:
                    p.data.add_(-group['weight_decay'], p.data)

                p.data.addcdiv_(-step_size, exp_avg, denom)

        return loss


class ExpAnnealLR(_LRScheduler):
    """Exponentially anneal the learning rate of each parameter group
    from the initial lr to end_lr over a number of iterations.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        end_lr (float): The final learning rate.
        iterations (int): The number of iterations over which to increase the
            learning rate.
        last_epoch (int): The index of the last iteration. Default: -1.
    """
    def __init__(self, optimizer, end_lr, iterations, last_epoch=-1):
        self.end_lr = end_lr
        self.iterations = iterations
        super(ExpAnnealLR, self).__init__(optimizer, last_epoch=last_epoch)

    def get_lr(self):
        iteration = self.last_epoch + 1
        pct = iteration / self.iterations
        return [base_lr * (self.end_lr / base_lr) ** pct for base_lr in self.base_lrs]


class ReduceLRWDOnPlateau(ReduceLROnPlateau):
    """Reduce learning rate and weight decay when a metric has stopped
    improving. Models often benefit from reducing the learning rate by
    a factor of 2-10 once learning stagnates. This scheduler reads a metric
    quantity and if no improvement is seen for a 'patience' number
    of epochs, the learning rate and weight decay factor is reduced for
    optimizers that implement the the weight decay method from the paper
    `Fixing Weight Decay Regularization in Adam`_.

    .. _Fixing Weight Decay Regularization in Adam:
        https://arxiv.org/abs/1711.05101

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        mode (str): One of `min`, `max`. In `min` mode, lr will
            be reduced when the quantity monitored has stopped
            decreasing; in `max` mode it will be reduced when the
            quantity monitored has stopped increasing. Default: 'min'.
        factor (float): Factor by which the learning rate will be
            reduced. new_lr = lr * factor. Default: 0.1.
        patience (int): Number of epochs with no improvement after
            which learning rate will be reduced. For example, if
            `patience = 2`, then we will ignore the first 2 epochs
            with no improvement, and will only decrease the LR after the
            3rd epoch if the loss still hasn't improved then.
            Default: 10.
        verbose (bool): If ``True``, prints a message to stdout for
            each update. Default: ``False``.
        threshold (float): Threshold for measuring the new optimum,
            to only focus on significant changes. Default: 1e-4.
        threshold_mode (str): One of `rel`, `abs`. In `rel` mode,
            dynamic_threshold = best * ( 1 + threshold ) in 'max'
            mode or best * ( 1 - threshold ) in `min` mode.
            In `abs` mode, dynamic_threshold = best + threshold in
            `max` mode or best - threshold in `min` mode. Default: 'rel'.
        cooldown (int): Number of epochs to wait before resuming
            normal operation after lr has been reduced. Default: 0.
        min_lr (float or list): A scalar or a list of scalars. A
            lower bound on the learning rate of all param groups
            or each group respectively. Default: 0.
        eps (float): Minimal decay applied to lr. If the difference
            between new and old lr is smaller than eps, the update is
            ignored. Default: 1e-8.

    Example:
        >>> optimizer = AdamW(model.parameters(), lr=0.1, weight_decay=1e-3)
        >>> scheduler = ReduceLRWDOnPlateau(optimizer, 'min')
        >>> for epoch in range(10):
        >>>     train(...)
        >>>     val_loss = validate(...)
        >>>     # Note that step should be called after validate()
        >>>     scheduler.step(val_loss)
    """
    def step(self, metrics, epoch=None):
        current = metrics
        if epoch is None:
            epoch = self.last_epoch = self.last_epoch + 1
        self.last_epoch = epoch

        if self.is_better(current, self.best):
            self.best = current
            self.num_bad_epochs = 0
        else:
            self.num_bad_epochs += 1

        if self.in_cooldown:
            self.cooldown_counter -= 1
            self.num_bad_epochs = 0  # ignore any bad epochs in cooldown

        if self.num_bad_epochs > self.patience:
            self._reduce_lr(epoch)
            self._reduce_weight_decay(epoch)
            self.cooldown_counter = self.cooldown
            self.num_bad_epochs = 0

    def _reduce_weight_decay(self, epoch):
        for i, param_group in enumerate(self.optimizer.param_groups):
            if param_group['weight_decay'] != 0:
                old_weight_decay = float(param_group['weight_decay'])
                new_weight_decay = max(old_weight_decay * self.factor, self.min_lrs[i])
                if old_weight_decay - new_weight_decay > self.eps:
                    param_group['weight_decay'] = new_weight_decay
                    if self.verbose:
                        log.info(f'Epoch {epoch}: reducing weight decay factor of group {i} to {new_weight_decay:.4e}.')



from .parameter import Parameter, SEQUENCE_TAGGER_PARAMETERS, TRAINING_PARAMETERS, DOCUMENT_EMBEDDING_PARAMETERS
from .param_selection import SequenceTaggerParamSelector, TextClassifierParamSelector, SearchSpace


import shutil
from pathlib import Path

import flair
from flair.data_fetcher import NLPTask, NLPTaskDataFetcher


def test_load_imdb_data(tasks_base_path):
    # get training, test and dev data
    corpus = NLPTaskDataFetcher.load_corpus(NLPTask.IMDB, tasks_base_path)

    assert len(corpus.train) == 5
    assert len(corpus.dev) == 5
    assert len(corpus.test) == 5


def test_load_ag_news_data(tasks_base_path):
    # get training, test and dev data
    corpus = NLPTaskDataFetcher.load_corpus(NLPTask.AG_NEWS, tasks_base_path)

    assert len(corpus.train) == 10
    assert len(corpus.dev) == 10
    assert len(corpus.test) == 10


def test_load_sequence_labeling_data(tasks_base_path):
    # get training, test and dev data
    corpus = NLPTaskDataFetcher.load_corpus(NLPTask.FASHION, tasks_base_path)

    assert len(corpus.train) == 6
    assert len(corpus.dev) == 1
    assert len(corpus.test) == 1


def test_load_germeval_data(tasks_base_path):
    # get training, test and dev data
    corpus = NLPTaskDataFetcher.load_corpus(NLPTask.GERMEVAL, tasks_base_path)

    assert len(corpus.train) == 2
    assert len(corpus.dev) == 1
    assert len(corpus.test) == 1


def test_load_ud_english_data(tasks_base_path):
    # get training, test and dev data
    corpus = NLPTaskDataFetcher.load_corpus(NLPTask.UD_ENGLISH, tasks_base_path)

    assert len(corpus.train) == 6
    assert len(corpus.test) == 4
    assert len(corpus.dev) == 2


def test_load_no_dev_data(tasks_base_path):
    # get training, test and dev data
    corpus = NLPTaskDataFetcher.load_column_corpus(tasks_base_path / 'fashion_nodev', {0: 'text', 2: 'ner'})

    assert len(corpus.train) == 5
    assert len(corpus.dev) == 1
    assert len(corpus.test) == 1


def test_load_no_dev_data_explicit(tasks_base_path):
    # get training, test and dev data
    corpus = NLPTaskDataFetcher.load_column_corpus(tasks_base_path / 'fashion_nodev', {0: 'text', 2: 'ner'}, train_file='train.tsv', test_file='test.tsv')

    assert len(corpus.train) == 5
    assert len(corpus.dev) == 1
    assert len(corpus.test) == 1

def test_multi_corpus(tasks_base_path):
    # get two corpora as one
    corpus = NLPTaskDataFetcher.load_corpora([NLPTask.FASHION, NLPTask.GERMEVAL], tasks_base_path)

    assert len(corpus.train) == 8
    assert len(corpus.dev) == 2
    assert len(corpus.test) == 2


def test_download_load_data(tasks_base_path):
    # get training, test and dev data for full English UD corpus from web
    corpus = NLPTaskDataFetcher.load_corpus(NLPTask.UD_ENGLISH)

    assert len(corpus.train) == 12543
    assert len(corpus.dev) == 2002
    assert len(corpus.test) == 2077

    # clean up data directory
    shutil.rmtree(Path(flair.file_utils.CACHE_ROOT) / 'datasets' / 'ud_english')



from sklearn.manifold import TSNE
import tqdm
import numpy


class _Transform:
    def __init__(self):
        pass

    def fit(self, X):
        return self.transform.fit_transform(X)


class tSNE(_Transform):
    def __init__(self):
        super().__init__()

        self.transform = \
            TSNE(n_components=2, verbose=1, perplexity=40, n_iter=300)


class Visualizer(object):

    def visualize_word_emeddings(self, embeddings, sentences, output_file):
        X = self.prepare_word_embeddings(embeddings, sentences)
        contexts = self.word_contexts(sentences)

        trans_ = tSNE()
        reduced = trans_.fit(X)

        self.visualize(reduced, contexts, output_file)

    def visualize_char_emeddings(self, embeddings, sentences, output_file):
        X = self.prepare_char_embeddings(embeddings, sentences)
        contexts = self.char_contexts(sentences)

        trans_ = tSNE()
        reduced = trans_.fit(X)

        self.visualize(reduced, contexts, output_file)

    @staticmethod
    def prepare_word_embeddings(embeddings, sentences):
        X = []

        for sentence in tqdm.tqdm(sentences):
            embeddings.embed(sentence)

            for i, token in enumerate(sentence):
                X.append(token.embedding.detach().numpy()[None, :])

        X = numpy.concatenate(X, 0)

        return X

    @staticmethod
    def word_contexts(sentences):
        contexts = []

        for sentence in sentences:

            strs = [x.text for x in sentence.tokens]

            for i, token in enumerate(strs):
                prop = '<b><font color="red"> {token} </font></b>'.format(
                    token=token)

                prop = ' '.join(strs[max(i - 4, 0):i]) + prop
                prop = prop + ' '.join(strs[i + 1:min(len(strs), i + 5)])

                contexts.append('<p>' + prop + '</p>')

        return contexts

    @staticmethod
    def prepare_char_embeddings(embeddings, sentences):
        X = []

        for sentence in tqdm.tqdm(sentences):

            sentence = ' '.join([x.text for x in sentence])

            hidden = embeddings.lm.get_representation([sentence])
            X.append(hidden.squeeze().detach().numpy())

        X = numpy.concatenate(X, 0)

        return X

    @staticmethod
    def char_contexts(sentences):
        contexts = []

        for sentence in sentences:
            sentence = ' '.join([token.text for token in sentence])

            for i, char in enumerate(sentence):

                context = '<span style="background-color: yellow"><b>{}</b></span>'.format(char)
                context = ''.join(sentence[max(i - 30, 0):i]) + context
                context = context + ''.join(sentence[i + 1:min(len(sentence), i + 30)])

                contexts.append(context)

        return contexts

    @staticmethod
    def visualize(X, contexts, file):
        import matplotlib.pyplot
        import mpld3

        fig, ax = matplotlib.pyplot.subplots()

        ax.grid(True, alpha=0.3)

        points = ax.plot(X[:, 0], X[:, 1], 'o', color='b',
                         mec='k', ms=5, mew=1, alpha=.6)

        ax.set_xlabel('x')
        ax.set_ylabel('y')
        ax.set_title('Hover mouse to reveal context', size=20)

        tooltip = mpld3.plugins.PointHTMLTooltip(
            points[0],
            contexts,
            voffset=10,
            hoffset=10
        )

        mpld3.plugins.connect(fig, tooltip)

        mpld3.save_html(fig, file)



import os

import pytest

from flair.visual import *
from flair.data import Sentence
from flair.embeddings import CharLMEmbeddings, StackedEmbeddings
import numpy

from flair.visual.manifold import Visualizer, tSNE
from flair.visual.training_curves import Plotter


@pytest.mark.slow
def test_visualize_word_emeddings(resources_path):

    with open(resources_path / 'visual/snippet.txt') as f:
        sentences = [x for x in f.read().split('\n') if x]

    sentences = [Sentence(x) for x in sentences]

    charlm_embedding_forward = CharLMEmbeddings('news-forward')
    charlm_embedding_backward = CharLMEmbeddings('news-backward')

    embeddings = StackedEmbeddings([charlm_embedding_backward, charlm_embedding_forward])

    visualizer = Visualizer()
    visualizer.visualize_word_emeddings(embeddings, sentences, str(resources_path / 'visual/sentence_embeddings.html'))

    # clean up directory
    (resources_path / 'visual/sentence_embeddings.html').unlink()


@pytest.mark.slow
def test_visualize_word_emeddings(resources_path):

    with open(resources_path / 'visual/snippet.txt') as f:
        sentences = [x for x in f.read().split('\n') if x]

    sentences = [Sentence(x) for x in sentences]

    charlm_embedding_forward = CharLMEmbeddings('news-forward')

    visualizer = Visualizer()
    visualizer.visualize_char_emeddings(charlm_embedding_forward, sentences, str(resources_path / 'visual/sentence_embeddings.html'))

    # clean up directory
    (resources_path / 'visual/sentence_embeddings.html').unlink()


@pytest.mark.slow
def test_visualize(resources_path):

    with open(resources_path / 'visual/snippet.txt') as f:
        sentences = [x for x in f.read().split('\n') if x]

    sentences = [Sentence(x) for x in sentences]

    embeddings = CharLMEmbeddings('news-forward')

    visualizer = Visualizer()

    X_forward = visualizer.prepare_char_embeddings(embeddings, sentences)

    embeddings = CharLMEmbeddings('news-backward')

    X_backward = visualizer.prepare_char_embeddings(embeddings, sentences)

    X = numpy.concatenate([X_forward, X_backward], axis=1)

    contexts = visualizer.char_contexts(sentences)

    trans_ = tSNE()
    reduced = trans_.fit(X)

    visualizer.visualize(reduced, contexts, str(resources_path / 'visual/char_embeddings.html'))

    # clean up directory
    (resources_path / 'visual/char_embeddings.html').unlink()


def test_highlighter(resources_path):
    with (resources_path / 'visual/snippet.txt').open() as f:
        sentences = [x for x in f.read().split('\n') if x]

    embeddings = CharLMEmbeddings('news-forward')

    features = embeddings.lm.get_representation(sentences[0]).squeeze()

    Highlighter().highlight_selection(features, sentences[0], n=1000, file_=str(resources_path / 'visual/highligh.html'))

    # clean up directory
    (resources_path / 'visual/highligh.html').unlink()


def test_plotting_training_curves_and_weights(resources_path):
    plotter = Plotter()
    plotter.plot_training_curves(resources_path / 'visual/loss.tsv')
    plotter.plot_weights(resources_path / 'visual/weights.txt')

    # clean up directory
    (resources_path / 'visual/weights.png').unlink()
    (resources_path / 'visual/training.png').unlink()



from typing import List, Dict, Union
import re
import logging
from enum import Enum
from pathlib import Path

import flair
from flair.data import Sentence, TaggedCorpus, Token, MultiCorpus
from flair.file_utils import cached_path

log = logging.getLogger('flair')


class NLPTask(Enum):
    # conll 2000 column format
    CONLL_2000 = 'conll_2000'

    # conll 03 NER column format
    CONLL_03 = 'conll_03'
    CONLL_03_GERMAN = 'conll_03_german'
    CONLL_03_DUTCH = 'conll_03_dutch'
    CONLL_03_SPANISH = 'conll_03_spanish'

    # WNUT-17
    WNUT_17 = 'wnut_17'

    # -- WikiNER datasets
    WIKINER_ENGLISH = 'wikiner_english'
    WIKINER_GERMAN = 'wikiner_german'
    WIKINER_FRENCH = 'wikiner_french'
    WIKINER_SPANISH = 'wikiner_spanish'
    WIKINER_ITALIAN = 'wikiner_italian'
    WIKINER_DUTCH = 'wikiner_dutch'
    WIKINER_POLISH = 'wikiner_polish'
    WIKINER_PORTUGUESE = 'wikiner_portuguese'
    WIKINER_RUSSIAN = 'wikiner_russian'

    # -- Universal Dependencies
    # Germanic
    UD_ENGLISH = 'ud_english'
    UD_GERMAN = 'ud_german'
    UD_DUTCH = 'ud_dutch'
    # Romance
    UD_FRENCH = 'ud_french'
    UD_ITALIAN = 'ud_italian'
    UD_SPANISH = 'ud_spanish'
    UD_PORTUGUESE = 'ud_portuguese'
    UD_ROMANIAN = 'ud_romanian'
    UD_CATALAN = 'ud_catalan'
    # West-Slavic
    UD_POLISH = 'ud_polish'
    UD_CZECH = 'ud_czech'
    UD_SLOVAK = 'ud_slovak'
    # South-Slavic
    UD_SLOVENIAN = 'ud_slovenian'
    UD_CROATIAN = 'ud_croatian'
    UD_SERBIAN = 'ud_serbian'
    UD_BULGARIAN = 'ud_bulgarian'
    # East-Slavic
    UD_RUSSIAN = 'ud_russian'
    # Scandinavian
    UD_SWEDISH = 'ud_swedish'
    UD_DANISH = 'ud_danish'
    UD_NORWEGIAN = 'ud_norwegian'
    UD_FINNISH = 'ud_finnish'
    # Asian
    UD_ARABIC = 'ud_arabic'
    UD_HEBREW = 'ud_hebrew'
    UD_TURKISH = 'ud_turkish'
    UD_PERSIAN = 'ud_persian'
    UD_HINDI = 'ud_hindi'
    UD_INDONESIAN = 'ud_indonesian'
    UD_JAPANESE = 'ud_japanese'
    UD_CHINESE = 'ud_chinese'
    UD_KOREAN = 'ud_korean'

    # other datasets
    ONTONER = 'ontoner'
    FASHION = 'fashion'
    GERMEVAL = 'germeval'
    SRL = 'srl'
    WSD = 'wsd'
    CONLL_12 = 'conll_12'
    PENN = 'penn'
    ONTONOTES = 'ontonotes'

    # text classification format
    IMDB = 'imdb'
    AG_NEWS = 'ag_news'


class NLPTaskDataFetcher:

    @staticmethod
    def load_corpora(tasks: List[Union[NLPTask, str]], base_path: Path = None) -> MultiCorpus:
        return MultiCorpus([NLPTaskDataFetcher.load_corpus(task, base_path) for task in tasks])

    @staticmethod
    def load_corpus(task: Union[NLPTask, str], base_path: [str, Path] = None) -> TaggedCorpus:
        """
        Helper function to fetch a TaggedCorpus for a specific NLPTask. For this to work you need to first download
        and put into the appropriate folder structure the corresponsing NLP task data. The tutorials on
        https://github.com/zalandoresearch/flair give more info on how to do this. Alternatively, you can use this
        code to create your own data fetchers.
        :param task: specification of the NLPTask you wish to get
        :param base_path: path to data folder containing tasks sub folders
        :return: a TaggedCorpus consisting of train, dev and test data
        """

        # first, try to fetch dataset online
        if type(task) is NLPTask:
            NLPTaskDataFetcher.download_dataset(task)

        # default dataset folder is the cache root
        if not base_path:
            base_path = Path(flair.file_utils.CACHE_ROOT) / 'datasets'

        if type(base_path) == str:
            base_path: Path = Path(base_path)

        # get string value if enum is passed
        task = task.value if type(task) is NLPTask else task

        data_folder = base_path / task.lower()

        # the CoNLL 2000 task on chunking has three columns: text, pos and np (chunk)
        if task == NLPTask.CONLL_2000.value:
            columns = {0: 'text', 1: 'pos', 2: 'np'}

            return NLPTaskDataFetcher.load_column_corpus(data_folder,
                                                         columns,
                                                         tag_to_biloes='np')

        # many NER tasks follow the CoNLL 03 format with four colulms: text, pos, np and ner tag
        if task == NLPTask.CONLL_03.value or task == NLPTask.ONTONER.value or task == NLPTask.FASHION.value:
            columns = {0: 'text', 1: 'pos', 2: 'np', 3: 'ner'}

            return NLPTaskDataFetcher.load_column_corpus(data_folder,
                                                         columns,
                                                         tag_to_biloes='ner',
                                                         )

        # the CoNLL 03 task for German has an additional lemma column
        if task == NLPTask.CONLL_03_GERMAN.value:
            columns = {0: 'text', 1: 'lemma', 2: 'pos', 3: 'np', 4: 'ner'}

            return NLPTaskDataFetcher.load_column_corpus(data_folder,
                                                         columns,
                                                         tag_to_biloes='ner')

        # the CoNLL 03 task for Dutch has no NP column
        if task == NLPTask.CONLL_03_DUTCH.value or task.startswith('wikiner'):
            columns = {0: 'text', 1: 'pos', 2: 'ner'}

            return NLPTaskDataFetcher.load_column_corpus(data_folder,
                                                         columns,
                                                         tag_to_biloes='ner')

        # the CoNLL 03 task for Spanish only has two columns
        if task == NLPTask.CONLL_03_SPANISH.value or task == NLPTask.WNUT_17.value:
            columns = {0: 'text', 1: 'ner'}

            return NLPTaskDataFetcher.load_column_corpus(data_folder,
                                                         columns,
                                                         tag_to_biloes='ner')

        # the GERMEVAL task only has two columns: text and ner
        if task == NLPTask.GERMEVAL.value:
            columns = {1: 'text', 2: 'ner'}

            return NLPTaskDataFetcher.load_column_corpus(data_folder,
                                                         columns,
                                                         tag_to_biloes='ner')

        # WSD tasks may be put into this column format
        if task == NLPTask.WSD.value:
            columns = {0: 'text', 1: 'lemma', 2: 'pos', 3: 'sense'}
            return NLPTaskDataFetcher.load_column_corpus(data_folder,
                                                         columns,
                                                         train_file='semcor.tsv',
                                                         test_file='semeval2015.tsv')

        # the UD corpora follow the CoNLL-U format, for which we have a special reader
        if task.startswith('ud_') or task in [NLPTask.ONTONOTES.value, NLPTask.CONLL_12.value, NLPTask.PENN.value]:
            return NLPTaskDataFetcher.load_ud_corpus(data_folder)

        # for text classifiers, we use our own special format
        if task == NLPTask.IMDB.value or task == NLPTask.AG_NEWS.value:
            return NLPTaskDataFetcher.load_classification_corpus(data_folder)

    @staticmethod
    def load_column_corpus(
            data_folder: Union[str, Path],
            column_format: Dict[int, str],
            train_file=None,
            test_file=None,
            dev_file=None,
            tag_to_biloes=None) -> TaggedCorpus:
        """
        Helper function to get a TaggedCorpus from CoNLL column-formatted task data such as CoNLL03 or CoNLL2000.

        :param data_folder: base folder with the task data
        :param column_format: a map specifying the column format
        :param train_file: the name of the train file
        :param test_file: the name of the test file
        :param dev_file: the name of the dev file, if None, dev data is sampled from train
        :param tag_to_biloes: whether to convert to BILOES tagging scheme
        :return: a TaggedCorpus with annotated train, dev and test data
        """

        if type(data_folder) == str:
            data_folder: Path = Path(data_folder)

        if train_file is not None:
            train_file = data_folder / train_file
        if test_file is not None:
            test_file = data_folder / test_file
        if dev_file is not None:
            dev_file = data_folder / dev_file

        # automatically identify train / test / dev files
        if train_file is None:
            for file in data_folder.iterdir():
                file_name = file.name
                if file_name.endswith('.gz'): continue
                if 'train' in file_name and not '54019' in file_name:
                    train_file = file
                if 'dev' in file_name:
                    dev_file = file
                if 'testa' in file_name:
                    dev_file = file
                if 'testb' in file_name:
                    test_file = file

            # if no test file is found, take any file with 'test' in name
            if test_file is None:
                for file in data_folder.iterdir():
                    file_name = file.name
                    if file_name.endswith('.gz'): continue
                    if 'test' in file_name:
                        test_file = file

        log.info("Reading data from {}".format(data_folder))
        log.info("Train: {}".format(train_file))
        log.info("Dev: {}".format(dev_file))
        log.info("Test: {}".format(test_file))

        # get train and test data
        sentences_train: List[Sentence] = NLPTaskDataFetcher.read_column_data(train_file, column_format)

        # read in test file if exists, otherwise sample 10% of train data as test dataset
        if test_file is not None:
            sentences_test: List[Sentence] = NLPTaskDataFetcher.read_column_data(test_file, column_format)
        else:
            sentences_test: List[Sentence] = [sentences_train[i] for i in
                                              NLPTaskDataFetcher.__sample(len(sentences_train), 0.1)]
            sentences_train = [x for x in sentences_train if x not in sentences_test]

        # read in dev file if exists, otherwise sample 10% of train data as dev dataset
        if dev_file is not None:
            sentences_dev: List[Sentence] = NLPTaskDataFetcher.read_column_data(dev_file, column_format)
        else:
            sentences_dev: List[Sentence] = [sentences_train[i] for i in
                                             NLPTaskDataFetcher.__sample(len(sentences_train), 0.1)]
            sentences_train = [x for x in sentences_train if x not in sentences_dev]

        if tag_to_biloes is not None:
            # convert tag scheme to iobes
            for sentence in sentences_train + sentences_test + sentences_dev:
                sentence: Sentence = sentence
                sentence.convert_tag_scheme(tag_type=tag_to_biloes, target_scheme='iobes')

        return TaggedCorpus(sentences_train, sentences_dev, sentences_test, name=data_folder.name)

    @staticmethod
    def load_ud_corpus(
            data_folder: Union[str,Path],
            train_file=None,
            test_file=None,
            dev_file=None) -> TaggedCorpus:
        """
        Helper function to get a TaggedCorpus from CoNLL-U column-formatted task data such as the UD corpora

        :param data_folder: base folder with the task data
        :param train_file: the name of the train file
        :param test_file: the name of the test file
        :param dev_file: the name of the dev file, if None, dev data is sampled from train
        :return: a TaggedCorpus with annotated train, dev and test data
        """
        # automatically identify train / test / dev files
        if train_file is None:
            for file in data_folder.iterdir():
                file_name = file.name
                if 'train' in file_name:
                    train_file = file
                if 'test' in file_name:
                    test_file = file
                if 'dev' in file_name:
                    dev_file = file
                if 'testa' in file_name:
                    dev_file = file
                if 'testb' in file_name:
                    test_file = file

        log.info("Reading data from {}".format(data_folder))
        log.info("Train: {}".format(train_file))
        log.info("Dev: {}".format(dev_file))
        log.info("Test: {}".format(test_file))

        sentences_train: List[Sentence] = NLPTaskDataFetcher.read_conll_ud(train_file)
        sentences_test: List[Sentence] = NLPTaskDataFetcher.read_conll_ud(test_file)
        sentences_dev: List[Sentence] = NLPTaskDataFetcher.read_conll_ud(dev_file)

        return TaggedCorpus(sentences_train, sentences_dev, sentences_test, name=data_folder.name)

    @staticmethod
    def load_classification_corpus(
            data_folder: Union[str,Path],
            train_file=None,
            test_file=None,
            dev_file=None) -> TaggedCorpus:
        """
        Helper function to get a TaggedCorpus from text classification-formatted task data

        :param data_folder: base folder with the task data
        :param train_file: the name of the train file
        :param test_file: the name of the test file
        :param dev_file: the name of the dev file, if None, dev data is sampled from train
        :return: a TaggedCorpus with annotated train, dev and test data
        """

        if type(data_folder) == str:
            data_folder: Path = Path(data_folder)

        if train_file is not None:
            train_file = data_folder / train_file
        if test_file is not None:
            test_file = data_folder / test_file
        if dev_file is not None:
            dev_file = data_folder / dev_file

        # automatically identify train / test / dev files
        if train_file is None:
            for file in data_folder.iterdir():
                file_name = file.name
                if 'train' in file_name:
                    train_file = file
                if 'test' in file_name:
                    test_file = file
                if 'dev' in file_name:
                    dev_file = file
                if 'testa' in file_name:
                    dev_file = file
                if 'testb' in file_name:
                    test_file = file

        log.info("Reading data from {}".format(data_folder))
        log.info("Train: {}".format(train_file))
        log.info("Dev: {}".format(dev_file))
        log.info("Test: {}".format(test_file))

        sentences_train: List[Sentence] = NLPTaskDataFetcher.read_text_classification_file(train_file)
        sentences_test: List[Sentence] = NLPTaskDataFetcher.read_text_classification_file(test_file)
        sentences_dev: List[Sentence] = NLPTaskDataFetcher.read_text_classification_file(dev_file)

        return TaggedCorpus(sentences_train, sentences_dev, sentences_test)

    @staticmethod
    def read_text_classification_file(path_to_file: Union[str,Path], max_tokens_per_doc=-1) -> List[Sentence]:
        """
        Reads a data file for text classification. The file should contain one document/text per line.
        The line should have the following format:
        __label__<class_name> <text>
        If you have a multi class task, you can have as many labels as you want at the beginning of the line, e.g.,
        __label__<class_name_1> __label__<class_name_2> <text>
        :param path_to_file: the path to the data file
        :param max_tokens_per_doc: Take only documents that contain number of tokens less or equal to this value. If
        set to -1 all documents are taken.
        :return: list of sentences
        """
        label_prefix = '__label__'
        sentences = []

        with open(str(path_to_file)) as f:
            for line in f:
                words = line.split()

                labels = []
                l_len = 0

                for i in range(len(words)):
                    if words[i].startswith(label_prefix):
                        l_len += len(words[i]) + 1
                        label = words[i].replace(label_prefix, "")
                        labels.append(label)
                    else:
                        break

                text = line[l_len:].strip()

                if text and labels:
                    sentence = Sentence(text, labels=labels, use_tokenizer=True)
                    if len(sentence) > max_tokens_per_doc:
                        sentence.tokens = sentence.tokens[:max_tokens_per_doc]
                    if len(sentence.tokens) > 0:
                        sentences.append(sentence)

        return sentences

    @staticmethod
    def read_column_data(path_to_column_file: Path,
                         column_name_map: Dict[int, str],
                         infer_whitespace_after: bool = True):
        """
        Reads a file in column format and produces a list of Sentence with tokenlevel annotation as specified in the
        column_name_map. For instance, by passing "{0: 'text', 1: 'pos', 2: 'np', 3: 'ner'}" as column_name_map you
        specify that the first column is the text (lexical value) of the token, the second the PoS tag, the third
        the chunk and the forth the NER tag.
        :param path_to_column_file: the path to the column file
        :param column_name_map: a map of column number to token annotation name
        :param infer_whitespace_after: if True, tries to infer whitespace_after field for Token
        :return: list of sentences
        """
        sentences: List[Sentence] = []

        try:
            lines: List[str] = open(str(path_to_column_file), encoding='utf-8').read().strip().split('\n')
        except:
            log.info('UTF-8 can\'t read: {} ... using "latin-1" instead.'.format(path_to_column_file))
            lines: List[str] = open(str(path_to_column_file), encoding='latin1').read().strip().split('\n')

        # most data sets have the token text in the first column, if not, pass 'text' as column
        text_column: int = 0
        for column in column_name_map:
            if column_name_map[column] == 'text':
                text_column = column

        sentence: Sentence = Sentence()
        for line in lines:

            if line.startswith('#'):
                continue

            if line.strip().replace('﻿', '') == '':
                if len(sentence) > 0:
                    sentence.infer_space_after()
                    sentences.append(sentence)
                sentence: Sentence = Sentence()

            else:
                fields: List[str] = re.split("\s+", line)
                token = Token(fields[text_column])
                for column in column_name_map:
                    if len(fields) > column:
                        if column != text_column:
                            token.add_tag(column_name_map[column], fields[column])

                sentence.add_token(token)

        if len(sentence.tokens) > 0:
            sentence.infer_space_after()
            sentences.append(sentence)

        return sentences

    @staticmethod
    def read_conll_ud(path_to_conll_file: Path) -> List[Sentence]:
        """
       Reads a file in CoNLL-U format and produces a list of Sentence with full morphosyntactic annotation
       :param path_to_conll_file: the path to the conll-u file
       :return: list of sentences
       """
        sentences: List[Sentence] = []

        lines: List[str] = open(path_to_conll_file, encoding='utf-8'). \
            read().strip().split('\n')

        sentence: Sentence = Sentence()
        for line in lines:

            fields: List[str] = re.split("\t+", line)
            if line == '':
                if len(sentence) > 0:
                    sentences.append(sentence)
                sentence: Sentence = Sentence()

            elif line.startswith('#'):
                continue
            elif '.' in fields[0]:
                continue
            elif '-' in fields[0]:
                continue
            else:
                token = Token(fields[1], head_id=int(fields[6]))
                token.add_tag('lemma', str(fields[2]))
                token.add_tag('upos', str(fields[3]))
                token.add_tag('pos', str(fields[4]))
                token.add_tag('dependency', str(fields[7]))

                for morph in str(fields[5]).split('|'):
                    if not "=" in morph: continue;
                    token.add_tag(morph.split('=')[0].lower(), morph.split('=')[1])

                if len(fields) > 10 and str(fields[10]) == 'Y':
                    token.add_tag('frame', str(fields[11]))

                sentence.add_token(token)

        if len(sentence.tokens) > 0: sentences.append(sentence)

        return sentences

    @staticmethod
    def __sample(total_number_of_sentences: int, percentage: float = 0.1) -> List[int]:
        import random
        sample_size: int = round(total_number_of_sentences * percentage)
        sample = random.sample(range(1, total_number_of_sentences), sample_size)
        return sample

    @staticmethod
    def download_dataset(task: NLPTask):

        # conll 2000 chunking task
        if task == NLPTask.CONLL_2000:
            conll_2000_path = 'https://www.clips.uantwerpen.be/conll2000/chunking/'
            data_file = Path(flair.file_utils.CACHE_ROOT) / 'datasets' / task.value / 'train.txt'
            if not data_file.is_file():
                cached_path(f'{conll_2000_path}train.txt.gz', Path('datasets') / task.value)
                cached_path(f'{conll_2000_path}test.txt.gz', Path('datasets') / task.value)
                import gzip, shutil
                with gzip.open(Path(flair.file_utils.CACHE_ROOT) / 'datasets' / task.value / 'train.txt.gz',
                               'rb') as f_in:
                    with open(Path(flair.file_utils.CACHE_ROOT) / 'datasets' / task.value / 'train.txt', 'wb') as f_out:
                        shutil.copyfileobj(f_in, f_out)
                with gzip.open(Path(flair.file_utils.CACHE_ROOT) / 'datasets' / task.value / 'test.txt.gz',
                               'rb') as f_in:
                    with open(Path(flair.file_utils.CACHE_ROOT) / 'datasets' / task.value / 'test.txt', 'wb') as f_out:
                        shutil.copyfileobj(f_in, f_out)

        if task == NLPTask.WNUT_17:
            wnut_path = 'https://noisy-text.github.io/2017/files/'
            cached_path(f'{wnut_path}wnut17train.conll', Path('datasets') / task.value)
            cached_path(f'{wnut_path}emerging.dev.conll', Path('datasets') / task.value)
            cached_path(f'{wnut_path}emerging.test.annotated', Path('datasets') / task.value)

        # Wikiner NER task
        wikiner_path = 'https://raw.githubusercontent.com/dice-group/FOX/master/input/Wikiner/'
        if task.value.startswith('wikiner'):
            lc = ''
            if task == NLPTask.WIKINER_ENGLISH: lc = 'en'
            if task == NLPTask.WIKINER_GERMAN: lc = 'de'
            if task == NLPTask.WIKINER_DUTCH: lc = 'nl'
            if task == NLPTask.WIKINER_FRENCH: lc = 'fr'
            if task == NLPTask.WIKINER_ITALIAN: lc = 'it'
            if task == NLPTask.WIKINER_SPANISH: lc = 'es'
            if task == NLPTask.WIKINER_PORTUGUESE: lc = 'pt'
            if task == NLPTask.WIKINER_POLISH: lc = 'pl'
            if task == NLPTask.WIKINER_RUSSIAN: lc = 'ru'

            data_file = Path(flair.file_utils.CACHE_ROOT) / 'datasets' / task.value / f'aij-wikiner-{lc}-wp3.train'
            if not data_file.is_file():

                cached_path(f'{wikiner_path}aij-wikiner-{lc}-wp3.bz2', Path('datasets') / task.value)
                import bz2, shutil

                # unpack and write out in CoNLL column-like format
                bz_file = bz2.BZ2File(
                    Path(flair.file_utils.CACHE_ROOT) / 'datasets' / task.value / f'aij-wikiner-{lc}-wp3.bz2', 'rb')
                with bz_file as f, open(
                        Path(flair.file_utils.CACHE_ROOT) / 'datasets' / task.value / f'aij-wikiner-{lc}-wp3.train',
                        'w') as out:
                    for line in f:
                        line = line.decode('utf-8')
                        words = line.split(' ')
                        for word in words:
                            out.write('\t'.join(word.split('|')) + '\n')

        # CoNLL 02/03 NER
        conll_02_path = 'https://www.clips.uantwerpen.be/conll2002/ner/data/'
        if task == NLPTask.CONLL_03_DUTCH:
            cached_path(f'{conll_02_path}ned.testa', Path('datasets') / task.value)
            cached_path(f'{conll_02_path}ned.testb', Path('datasets') / task.value)
            cached_path(f'{conll_02_path}ned.train', Path('datasets') / task.value)
        if task == NLPTask.CONLL_03_SPANISH:
            cached_path(f'{conll_02_path}esp.testa', Path('datasets') / task.value)
            cached_path(f'{conll_02_path}esp.testb', Path('datasets') / task.value)
            cached_path(f'{conll_02_path}esp.train', Path('datasets') / task.value)

        # universal dependencies
        ud_path = 'https://raw.githubusercontent.com/UniversalDependencies/'
        # --- UD Germanic
        if task == NLPTask.UD_ENGLISH:
            cached_path(f'{ud_path}UD_English-EWT/master/en_ewt-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_English-EWT/master/en_ewt-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_English-EWT/master/en_ewt-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_GERMAN:
            cached_path(f'{ud_path}UD_German-GSD/master/de_gsd-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_German-GSD/master/de_gsd-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_German-GSD/master/de_gsd-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_DUTCH:
            cached_path(f'{ud_path}UD_Dutch-Alpino/master/nl_alpino-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Dutch-Alpino/master/nl_alpino-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Dutch-Alpino/master/nl_alpino-ud-train.conllu', Path('datasets') / task.value)

        # --- UD Romance
        if task == NLPTask.UD_FRENCH:
            cached_path(f'{ud_path}UD_French-GSD/master/fr_gsd-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_French-GSD/master/fr_gsd-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_French-GSD/master/fr_gsd-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_ITALIAN:
            cached_path(f'{ud_path}UD_Italian-ISDT/master/it_isdt-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Italian-ISDT/master/it_isdt-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Italian-ISDT/master/it_isdt-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_SPANISH:
            cached_path(f'{ud_path}UD_Spanish-GSD/master/es_gsd-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Spanish-GSD/master/es_gsd-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Spanish-GSD/master/es_gsd-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_PORTUGUESE:
            cached_path(f'{ud_path}UD_Portuguese-Bosque/blob/master/pt_bosque-ud-dev.conllu',
                        Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Portuguese-Bosque/blob/master/pt_bosque-ud-test.conllu',
                        Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Portuguese-Bosque/blob/master/pt_bosque-ud-train.conllu',
                        Path('datasets') / task.value)

        if task == NLPTask.UD_ROMANIAN:
            cached_path(f'{ud_path}UD_Romanian-RRT/master/ro_rrt-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Romanian-RRT/master/ro_rrt-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Romanian-RRT/master/ro_rrt-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_CATALAN:
            cached_path(f'{ud_path}UD_Catalan-AnCora/master/ca_ancora-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Catalan-AnCora/master/ca_ancora-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Catalan-AnCora/master/ca_ancora-ud-train.conllu', Path('datasets') / task.value)

        # --- UD West-Slavic
        if task == NLPTask.UD_POLISH:
            cached_path(f'{ud_path}UD_Polish-LFG/master/pl_lfg-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Polish-LFG/master/pl_lfg-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Polish-LFG/master/pl_lfg-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_CZECH:
            cached_path(f'{ud_path}UD_Czech-PDT/master/cs_pdt-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Czech-PDT/master/cs_pdt-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Czech-PDT/master/cs_pdt-ud-train-l.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_SLOVAK:
            cached_path(f'{ud_path}UD_Slovak-SNK/master/sk_snk-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Slovak-SNK/master/sk_snk-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Slovak-SNK/master/sk_snk-ud-train.conllu', Path('datasets') / task.value)

        # --- UD Scandinavian
        if task == NLPTask.UD_SWEDISH:
            cached_path(f'{ud_path}UD_Swedish-Talbanken/master/sv_talbanken-ud-dev.conllu',
                        Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Swedish-Talbanken/master/sv_talbanken-ud-test.conllu',
                        Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Swedish-Talbanken/master/sv_talbanken-ud-train.conllu',
                        Path('datasets') / task.value)

        if task == NLPTask.UD_DANISH:
            cached_path(f'{ud_path}UD_Danish-DDT/master/da_ddt-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Danish-DDT/master/da_ddt-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Danish-DDT/master/da_ddt-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_NORWEGIAN:
            cached_path(f'{ud_path}UD_Norwegian-Bokmaal/master/no_bokmaal-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Norwegian-Bokmaal/master/no_bokmaal-ud-test.conllu',
                        Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Norwegian-Bokmaal/master/no_bokmaal-ud-train.conllu',
                        Path('datasets') / task.value)

        if task == NLPTask.UD_FINNISH:
            cached_path(f'{ud_path}UD_Finnish-TDT/master/fi_tdt-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Finnish-TDT/master/fi_tdt-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Finnish-TDT/master/fi_tdt-ud-train.conllu', Path('datasets') / task.value)

        # --- UD South-Slavic
        if task == NLPTask.UD_SLOVENIAN:
            cached_path(f'{ud_path}UD_Slovenian-SSJ/master/sl_ssj-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Slovenian-SSJ/master/sl_ssj-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Slovenian-SSJ/master/sl_ssj-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_CROATIAN:
            cached_path(f'{ud_path}UD_Croatian-SET/master/hr_set-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Croatian-SET/master/hr_set-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Croatian-SET/master/hr_set-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_SERBIAN:
            cached_path(f'{ud_path}UD_Serbian-SET/master/sr_set-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Serbian-SET/master/sr_set-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Serbian-SET/master/sr_set-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_BULGARIAN:
            cached_path(f'{ud_path}UD_Bulgarian-BTB/master/bg_btb-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Bulgarian-BTB/master/bg_btb-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Bulgarian-BTB/master/bg_btb-ud-train.conllu', Path('datasets') / task.value)

        # --- UD Asian
        if task == NLPTask.UD_ARABIC:
            cached_path(f'{ud_path}UD_Arabic-PADT/master/ar_padt-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Arabic-PADT/master/ar_padt-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Arabic-PADT/master/ar_padt-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_HEBREW:
            cached_path(f'{ud_path}UD_Hebrew-HTB/master/he_htb-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Hebrew-HTB/master/he_htb-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Hebrew-HTB/master/he_htb-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_TURKISH:
            cached_path(f'{ud_path}UD_Turkish-IMST/master/tr_imst-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Turkish-IMST/master/tr_imst-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Turkish-IMST/master/tr_imst-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_PERSIAN:
            cached_path(f'{ud_path}UD_Persian-Seraji/master/fa_seraji-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Persian-Seraji/master/fa_seraji-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Persian-Seraji/master/fa_seraji-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_RUSSIAN:
            cached_path(f'{ud_path}UD_Russian-SynTagRus/master/ru_syntagrus-ud-dev.conllu',
                        Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Russian-SynTagRus/master/ru_syntagrus-ud-test.conllu',
                        Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Russian-SynTagRus/master/ru_syntagrus-ud-train.conllu',
                        Path('datasets') / task.value)

        if task == NLPTask.UD_HINDI:
            cached_path(f'{ud_path}UD_Hindi-HDTB/master/hi_hdtb-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Hindi-HDTB/master/hi_hdtb-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Hindi-HDTB/master/hi_hdtb-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_INDONESIAN:
            cached_path(f'{ud_path}UD_Indonesian-GSD/master/id_gsd-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Indonesian-GSD/master/id_gsd-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Indonesian-GSD/master/id_gsd-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_JAPANESE:
            cached_path(f'{ud_path}UD_Japanese-GSD/master/ja_gsd-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Japanese-GSD/master/ja_gsd-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Japanese-GSD/master/ja_gsd-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_CHINESE:
            cached_path(f'{ud_path}UD_Chinese-GSD/master/zh_gsd-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Chinese-GSD/master/zh_gsd-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Chinese-GSD/master/zh_gsd-ud-train.conllu', Path('datasets') / task.value)

        if task == NLPTask.UD_KOREAN:
            cached_path(f'{ud_path}UD_Korean-Kaist/master/ko_kaist-ud-dev.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Korean-Kaist/master/ko_kaist-ud-test.conllu', Path('datasets') / task.value)
            cached_path(f'{ud_path}UD_Korean-Kaist/master/ko_kaist-ud-train.conllu', Path('datasets') / task.value)



import shutil

from flair.data import Dictionary
from flair.trainers.language_model_trainer import TextCorpus


def test_train_resume_language_model_training(resources_path, results_base_path, tasks_base_path):
    # get default dictionary
    dictionary: Dictionary = Dictionary.load('chars')

    # get the example corpus and process at character level in forward direction
    corpus: TextCorpus = TextCorpus(resources_path / 'corpora/lorem_ipsum',
                                dictionary,
                                forward=True,
                                character_level=True)

    assert (corpus.test is not None)
    assert (corpus.train_files is not None)
    assert (corpus.valid is not None)
    assert (len(corpus.train_files) == 2)
